"""
Some codes are borrowed from https://github.com/PrefectHQ/prefect/blob/master/src/prefect/engine/runner.py
"""
import asyncio
import functools
import inspect
import threading
import traceback
from typing import Callable, List, Union, Optional, Coroutine, Set, Tuple

import flowsaber
from flowsaber.client.client import Client
from flowsaber.core.base import Component, enter_context
from flowsaber.core.utility.state import State, Failure, Scheduled, Cancelling, Cancelled
from flowsaber.server.database.models import RunInput, RunLogsInput, RunLogInput, FlowRunInput, TaskRunInput
from flowsaber.utility.utils import interrupt_main

AsyncFunc = Callable[..., Coroutine]


class RunException(RuntimeError):
    def __init__(self, *args, state: State = None):
        super().__init__(*args)
        self.state = state


def redirect_std_to_logger(method: Callable[..., State]) -> Callable[..., State]:
    @functools.wraps(method)
    def redirect(self: 'Runner', *args, **kwargs) -> State:
        import logging
        from contextlib import redirect_stdout, redirect_stderr, nullcontext
        from flowsaber.utility.logging import RedirectToLog

        rd_stdout_context = nullcontext()
        rd_stderr_context = nullcontext()
        cur_logger = flowsaber.context.logger
        updated_config = flowsaber.context.get(self.component.config_name, {})
        if updated_config.get('log_stdout', False):
            rd_stdout_context = redirect_stdout(RedirectToLog(cur_logger, level=logging.INFO))
        if updated_config.get('log_stderr', False):
            rd_stderr_context = redirect_stderr(RedirectToLog(cur_logger, level=logging.ERROR))

        with rd_stdout_context, rd_stderr_context:
            return method(self, *args, **kwargs)

    return redirect


def call_state_change_handlers(method: Callable[..., State]) -> Callable[..., State]:
    """A decorator checks the difference between _input and _output state of the wrapped method, if two states are
    not identical, trigger runner's handle_state_change method for calling all state change handlers.

    Parameters
    ----------
    method

    Returns
    -------

    """

    @functools.wraps(method)
    def check_and_run(self: "Runner", state: State = None, *args, **kwargs) -> State:
        new_state = method(self, state, *args, **kwargs)

        if new_state is not state:
            new_state = self.handle_state_change(state, new_state)

        return new_state

    return check_and_run


def catch_to_failure(method: Callable[..., State]) -> Callable[..., State]:
    """A decorator that wraps method into a method that automatically capture exceptions into Failure state.

    Parameters
    ----------
    method

    Returns
    -------

    """

    @functools.wraps(method)
    def catch_exception_to_failure(self: "Runner", *args, **kwargs) -> State:
        try:
            new_state = method(self, *args, **kwargs)
        # TODO how to handle this only once?
        # must use BaseException, because KeyBoardInterrupt is not Exception
        except BaseException as exc:
            tb = traceback.format_exc()
            if isinstance(exc, KeyboardInterrupt):
                e_str = f"{type(self).__name__} is cancelled with error: {exc}."
                new_state = Cancelled(result=exc, message=e_str, trace_back=tb)
            else:
                e_str = f"{type(self).__name__} meet Unexpected error: {exc} when calling method: {method} " \
                        f"with traceback: {tb}."
                new_state = Failure(result=exc, message=e_str, trace_back=tb)

        return new_state

    return catch_exception_to_failure


def run_timeout_signal(timeout: int, func: Callable, *args, **kwargs):
    """Run function in main thread in unix system with timeout using SIGALARM signal.

    References
    ----------
    https://github.com/pnpnpn/timeout-decorator/blob/master/timeout_decorator/timeout_decorator.py

    Parameters
    ----------
    func
    timeout
    args
    kwargs

    Returns
    -------

    """
    import signal

    def error_handler(signum, frame):
        raise TimeoutError(f"Execution timout out of {timeout}")

    old = signal.signal(signal.SIGALRM, error_handler)
    signal.setitimer(signal.ITIMER_REAL, timeout)
    try:

        # Raise the alarm if `timeout` seconds pass
        flowsaber.context.logger.debug(f"Sending alarm with {timeout}s timeout...")
        signal.alarm(timeout)
        flowsaber.context.logger.debug("Executing function in main thread...")
        res = func(*args, **kwargs)

    finally:
        signal.setitimer(signal.ITIMER_REAL, 0)
        signal.signal(signal.SIGALRM, old)

    return res


def run_timeout_thread(timeout: int, *args, **kwargs):
    """Run the task within timeout within a thread pool. Note that the flowsaber.context would be corrupted
    in the new thread.

    Parameters
    ----------
    func
    timeout
    kwargs

    Returns
    -------

    """
    from concurrent.futures import ThreadPoolExecutor

    try:
        with ThreadPoolExecutor(max_workers=1) as executor:
            fut = executor.submit(*args, **kwargs)
            return fut.result(timeout=timeout)
    finally:
        if not fut.done():
            fut.cancel()


class RunnerExecuteError(Exception):
    def __init__(self, *args, futures=None):
        super().__init__(*args)
        self.futures = futures or []


class RunnerExecutor(threading.Thread):
    class DoneException(Exception):
        pass

    def __init__(self, context=dict, **kwargs):
        super().__init__(**kwargs)
        self.context = context
        self.tasks: List[Callable[..., Tuple[Coroutine, Callable]]] = []
        self.task_done_callbacks: List[AsyncFunc] = []
        # running attributes
        self.futures: List[asyncio.Future] = []
        self.jobs: List[Tuple[Coroutine, Callable]] = []
        self.loop: Optional[asyncio.AbstractEventLoop] = None
        self.started: Optional[threading.Event] = None
        self.exception: Optional[BaseException] = None

    def start(self) -> None:
        self.started = threading.Event()
        super().start()
        self.started.wait()

    def siganl_kill(self):
        if self.started and self.started.is_set() and len(self.futures):
            # the first task is a stop future
            self.loop.call_soon_threadsafe(self.futures[0].set_exception, self.DoneException())

    def signal_stop(self):
        if self.started and self.started.is_set() and len(self.futures):
            # the first task is a stop future
            self.loop.call_soon_threadsafe(self.futures[0].set_result, None)

    def join(self, timeout: Optional[float] = ...) -> None:
        self.signal_stop()
        for _, stop in self.jobs:
            stop()
        super().join()
        self.started = None
        if self.exception:
            exception = self.exception
            self.exception = None
            raise exception

    async def main_loop(self):
        """Endlessly fetch async task anc schedule for running in asyncio envent loop untill some task raise an
        exception.

        Returns
        -------

        """

        def raise_on_error(loop, context):
            loop.default_exception_handler(context)
            self.siganl_kill()

        self.loop = asyncio.get_running_loop()
        self.loop.set_exception_handler(raise_on_error)
        self.futures.clear()
        self.jobs.clear()
        # add a stop future as stop signal of executor
        stop_future = asyncio.Future()
        self.futures.append(stop_future)

        while len(self.tasks):
            task_func = self.tasks.pop(-1)
            job = task_func()
            if isinstance(job, (list, tuple)):
                coro, stop, *_ = job
            else:
                coro, stop = job, lambda: 1
            assert inspect.iscoroutine(coro) and callable(stop) and not inspect.iscoroutinefunction(stop)

            fut = asyncio.create_task(coro)
            self.futures.append(fut)
            self.jobs.append((coro, stop))

        # now thread.start end blocking
        self.started.set()
        # use the stop_future to force cancel all tasks in case of Exception
        done, pending = await asyncio.wait(self.futures, return_when=asyncio.FIRST_EXCEPTION)
        # try to cancel all jobs
        for job in pending:
            job.cancel()

        # do callbacks
        for callback in self.task_done_callbacks:
            res = callback()
            if inspect.iscoroutine(res):
                await res

        # raise exception jobs
        error_futures = [task for task in done
                         if not task.cancelled()
                         and task.exception()
                         and task is not stop_future]
        futures = list(done) + list(pending)
        futures.remove(stop_future)
        try:
            if len(error_futures):
                first_exception = error_futures[0].exception()
                raise RunnerExecuteError(str(first_exception), futures=futures) from first_exception

            return futures
        finally:
            self.started.clear()
            self.futures.clear()
            self.jobs.clear()
            self.loop = None

    @enter_context
    def run(self):
        try:
            asyncio.run(self.main_loop())
        except BaseException as exc:
            self.exception = exc

    def add_task(self, task: Callable[..., Coroutine]):
        assert not self.started or not self.started.is_set(), "Can only add task before running"
        assert callable(task) and not inspect.iscoroutinefunction(task)
        self.tasks.append(task)


class Runner(object):
    """Base runner class, intended to be the state manager of runnable object like flow and task.

    Users need to add state change handlers in order to be informed when meeting state changes of some method.
    Methods of runner should both accept and return state, and need to be decorated with
    `call_state_change_handlers` decorator.
    """

    def __init__(self, server_address: str = None, id: str = None, name: str = None, labels: list = None, **kwargs):
        self.id = id or flowsaber.context.random_id
        self.name = name or flowsaber.context.random_id
        self.labels = labels or []
        self.component: Optional[Component] = None
        # used for executing background async tasks
        self.tasks: List[Callable[..., Coroutine]] = []
        self.executor: Optional[RunnerExecutor] = None
        self.state_change_handlers: Set[Callable] = set()
        # client
        self.server_address: str = server_address
        self.client: Optional[Client] = Client(self.server_address) if self.server_address else None
        # add tasks and state handlers
        self.state_change_handlers.add(self.logging_run_state)
        if self.client:
            self.add_task(self.update_run_state)
            # TODO may meet situations when update_run_state is not done but send_logs is done
            self.add_task(self.send_logs)

    def add_task(self, task: Callable[..., Tuple[Coroutine, Callable]]):
        # TODO each async task should add a exit async function
        assert callable(task) and not inspect.iscoroutinefunction(task), \
            "The task must be a function return a coroutine"
        self.tasks.append(task)

    def remove_async_tasks(self, task: Callable):
        self.tasks.remove(task)

    @property
    def context(self) -> dict:
        assert self.component is not None, 'component is not settled'
        return self.component.context

    @property
    def config(self) -> dict:
        assert self.component is not None, 'component is not settled'
        return self.component.config_dict

    @call_state_change_handlers
    def initialize_run(self, state: Union[State, None], **kwargs) -> State:
        if state is None:
            state = Scheduled()
        return state

    def handle_state_change(self, prev_state, cur_state):
        """Call all registered state change handlers with parameter of old_state and new_state.

        Parameters
        ----------
        prev_state
        cur_state

        Returns
        -------

        """
        handler = None
        try:
            for handler in self.state_change_handlers:
                if handler is None:
                    continue
                cur_state = handler(self, prev_state, cur_state) or cur_state
        except Exception as exc:
            e_str = f"Unexpected error: {exc} when calling state_handler: {handler}"
            cur_state = Failure(result=exc, message=e_str)
        return cur_state

    def run(self, state: State = None, **kwargs) -> State:
        try:
            kwargs.setdefault('context', {})
            self.enter_run(state, **kwargs)
            final_state = self.start_run(state=state, **kwargs)
            # for debug
            if 'flow' in type(self).__name__.lower() and isinstance(final_state, Failure):
                raise final_state.result
            return final_state
        finally:
            self.leave_run()

    def enter_run(self, *args, **kwargs):
        self.initialize_context(*args, **kwargs)
        # TODO to may redundant update, should move in the future
        with flowsaber.context(self.context):
            with flowsaber.context(kwargs.get('context', {})) as context:
                context_dict = context.to_dict()
        if self.client:
            self.executor = RunnerExecutor(context=context_dict)
            for task in self.tasks:
                self.executor.add_task(task)
            self.executor.task_done_callbacks.append(self.client.close)
            self.executor.start()

    def initialize_context(self, *args, **kwargs):
        pass

    def start_run(self, state: State = None, **kwargs) -> State:
        raise NotImplementedError

    def leave_run(self, *args, **kwargs):
        if self.client:
            self.executor.join()
            self.executor = None

    @call_state_change_handlers
    def set_state(self, old_state: State, state_type: type):
        assert issubclass(state_type, State)
        return state_type.copy(old_state)

    def serialize(self, state: State, state_only=True) -> RunInput:
        raise NotImplementedError

    @staticmethod
    def logging_run_state(runner: "Runner", old_state, new_state):
        flowsaber.context.logger.info(f"State change from [{old_state}] to [{new_state}]")

    def send_logs(self) -> Tuple[Coroutine, Callable]:
        """This should be ran in runner.executor's async main loop"""

        def enqueue_log(record):
            from datetime import datetime
            record_dict = {
                'level': record.levelname,
                'time': datetime.fromtimestamp(record.created).timestamp(),
                'task_id': getattr(record, 'task_id', None),
                'flow_id': getattr(record, 'flow_id', None),
                'taskrun_id': getattr(record, 'taskrun_id', None),
                'agent_id': getattr(record, 'agent_id', None),
                'message': record.message
            }
            run_log = RunLogInput(**record_dict)
            loop.call_soon_threadsafe(log_queue.put_nowait, run_log)

        client = self.client
        log_queue = asyncio.Queue()
        loop = asyncio.get_running_loop()
        # TODO not thread safe
        # add handler
        log_handler = flowsaber.log_handler
        log_handler.handler = enqueue_log

        async def main_loop():
            try:
                while True:
                    log = await log_queue.get()
                    if log is None:
                        break
                    fields = "success\n" \
                             "error"
                    await client.mutation('write_runlogs', RunLogsInput(logs=[log]), fields)
            finally:
                # TODO not thread safe
                # remove handler
                log_handler.handler = None

        def stop():
            if loop.is_running():
                loop.call_soon_threadsafe(log_queue.put_nowait, None)

        return main_loop(), stop

    def update_run_state(self) -> Tuple[Coroutine, Callable]:
        """Does not use create_task as these jobs has a strict order

        Parameters
        ----------
        runner

        Returns
        -------

        """

        def state_change_handler(runner: "Runner", old_state: State, new_state: State):
            async def update_run():
                await client.mutation(method, run_input, field='id')

            is_flow = 'flow' in type(runner).__name__.lower()
            method = 'update_flowrun' if is_flow else "update_taskrun"
            state_only = not (old_state is None and isinstance(new_state, Scheduled))
            run_input = runner.serialize(new_state, state_only=state_only)
            client = runner.client
            loop.call_soon_threadsafe(task_queue.put_nowait, update_run())

        loop = asyncio.get_running_loop()
        task_queue = asyncio.Queue()
        # TODO not thread safe
        # add handler
        self.state_change_handlers.add(state_change_handler)

        async def main_loop():
            try:
                while True:
                    coro = await task_queue.get()
                    if coro is None:
                        break
                    await coro
            finally:
                # TODO not thread safe
                # remove handler
                self.state_change_handlers.remove(state_change_handler)

        def stop():
            if loop.is_running():
                loop.call_soon_threadsafe(task_queue.put_nowait, None)

        return main_loop(), stop

    def check_cancellation(self):
        loop = asyncio.get_running_loop()
        client = self.client
        check_interval = self.config.get('check_cancellation_interval', 5)
        runner_id = self.id
        if 'flow' in type(self).__name__.lower():
            method = 'get_flowrun'
        else:
            method = 'get_taskrun'
        need_stop = asyncio.Event()

        async def main_loop():
            while not need_stop.is_set():
                await asyncio.sleep(check_interval)
                run = await client.query(method, input=runner_id, field="state { state_type }")
                state = State.from_dict(run['state'])
                if isinstance(state, Cancelling):
                    flowsaber.context.logger.error("Fetched flow cancelling request from server.")
                    interrupt_main()

        def stop():
            if loop.is_running():
                loop.call_soon_threadsafe(need_stop.set)

        return main_loop(), stop

    def maintain_heartbeat(self) -> Tuple[Coroutine, Callable]:
        client = self.client
        loop = asyncio.get_running_loop()

        need_stop = asyncio.Event()
        is_flow = 'flow' in type(self).__name__.lower()
        method = 'update_flowrun' if is_flow else "update_taskrun"
        runner_id = self.id

        async def main_loop():
            while not need_stop.is_set():
                await asyncio.sleep(5)
                run_input = FlowRunInput(id=runner_id) if is_flow else TaskRunInput(id=runner_id)
                await client.mutation(method, input=run_input, field="id")

        def stop():
            if loop.is_running():
                loop.call_soon_threadsafe(need_stop.set)

        return main_loop(), stop



import asyncio
import inspect
from collections import abc
from collections import deque
from typing import Union, Sequence, Optional, List

import flowsaber
from flowsaber.core.utility.target import END, End
from flowsaber.server.database.models import ChannelInput


class Fetcher(object):
    """Fetch simple provide a for/async for method support for classes implemented with get/get_nowait methods.
    The end of __next__/__anext__ is triggered by the appearance of END fetched from get/get_nowait.
    """

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self, k):
                setattr(self, k, v)

    def __aiter__(self):
        return self

    async def __anext__(self):
        value = await self.get()
        if isinstance(value, End):
            raise StopAsyncIteration
        else:
            return value

    def __iter__(self):
        return self

    def __next__(self):
        value = self.get_nowait()
        if isinstance(value, End):
            raise StopIteration
        else:
            return value

    async def get(self):
        return self.get_nowait()

    def get_nowait(self):
        raise NotImplementedError


class ConstantQueue(object):
    """A async queue will emit it's internal value infinitely. Like ordinary queue, the first element
    needs to be enqueued before fetching"""
    NOTSET = object()

    def __init__(self):
        self.value = self.NOTSET
        self.has_value = asyncio.Event()

    def put_nowait(self, item):
        if not self.has_value.is_set():
            self.has_value.set()
        self.value = item

    async def put(self, item):
        self.put_nowait(item)

    def get_nowait(self):
        if self.value is self.NOTSET:
            raise RuntimeError("The ConstantQueue is not initialized, please use ch.put/ch.put_nowait "
                               "to set the initial value")
        return self.value

    async def get(self):
        await self.has_value.wait()
        return self.get_nowait()

    def empty(self):
        return self.value is not self.NOTSET


class LazyAsyncQueue(Fetcher):
    """Internally it's an asyncio.Queue, but the inner queue will only be created when meets the first call
    it's fetcher/putter methods. It's designed like this to handler pickle or coroutine-loop problems.

    """

    def __init__(self, ch, queue_factory, **kwargs):
        super().__init__(**kwargs)
        self.ch: Channel = ch
        self.queue_factory = queue_factory
        self.queue: Optional[Union[asyncio.Queue, ConstantQueue]] = None

    def initialize_queue(self):
        if self.queue is None:
            self.queue = self.queue_factory()

    async def get(self):
        self.initialize_queue()
        if not self.ch.initialized:
            self.ch.initialize()
        return await self.queue.get()

    def get_nowait(self):
        self.initialize_queue()
        return self.queue.get_nowait()

    def put_nowait(self, item):
        self.initialize_queue()
        return self.queue.put_nowait(item)

    async def put(self, item):
        self.initialize_queue()
        if not self.ch.initialized:
            self.ch.initialize()
        return await self.queue.put(item)

    def empty(self):
        return self.queue.empty()


class ChannelBase(object):
    """A channel it self is an object for storing items by using put/put_nowait. To consume data from the channel,
    users must call ChannelBase.create_queue method to get a fetcher LazyAsyncQueue, and then call it's get/get_nowait method
    to fetched the data emitted by the channel.
    """

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self, k):
                setattr(self, k, v)

    def put_nowait(self, item):
        raise NotImplementedError

    async def put(self, item):
        return self.put_nowait(item)

    def create_queue(self) -> LazyAsyncQueue:
        raise NotImplementedError

    def __lshift__(self, other):
        """
        ch << 1 == ch.put_nowait(1)
        """
        self.put_nowait(other)
        return self

    def __rshift__(self, tasks) -> Union['Channel', Sequence['Channel']]:
        """
        ch >> task                   -> task(ch)
        ch >> [task1, _tasks, task3]  -> [task1(ch), task2(ch), task3(ch)]
        """
        if not isinstance(tasks, abc.Sequence):
            return tasks(self)

        outputs = [task(self) for task in tasks]
        if isinstance(tasks, tuple):
            outputs = tuple(tasks)
        return outputs

    def __or__(self, tasks) -> Union['Channel', Sequence['Channel']]:
        """
        ch | [a, b, c, d] equals to ch >> [a, b, c, d]
        """
        return self >> tasks

    @staticmethod
    def value(value, **kwargs):
        """
        Channel._output(1)
        """
        if callable(value):
            raise ValueError("You has passed a callable object as inputs, "
                             "you should explicitly specify the argument name like:"
                             "`ch.map(by=lambda x : x)`.")
        ch = ConstantChannel(**kwargs)
        ch.put_nowait(value)
        return ch

    @staticmethod
    def end():
        return ConstantChannel()

    @staticmethod
    def values(*args):
        """
        Channel.values(1, 2, 3, 4, 5)
        QueueChannel created by this method will always include a END signal
        """
        ch = Channel()
        for item in args:
            ch.put_nowait(item)
        ch.put_nowait(END)
        return ch

    @staticmethod
    def from_list(items: Sequence):
        """
        Channel.from_list([1, 2, 3, 4, 5])
        QueueChannel created by this method will always include a END signal
        """
        return Channel.values(*items)


class Channel(ChannelBase):
    """Subclass of ChannelBase implemented create_queue method, the mechanism for sending data to all created queue is
    simple, it just loop over all queues and push the item specified by call of put/put_nowait into all queues.
    Furthermore, to make LazyAsyncQueue initialized only in a running event loop, Channel uses a buffer to buffer all
    items pushed before entering the event loop.
    """

    def __init__(self, queue_factory: type = asyncio.Queue, **kwargs):
        super().__init__(**kwargs)
        self.buffer: deque = deque()
        self.initialized = False
        self.queues: List[LazyAsyncQueue] = []
        self.queue_factory = queue_factory
        # context info
        self.id = flowsaber.context.random_id
        self.task_id = flowsaber.context.get('task_id', None)
        self.flow_id = flowsaber.context.get('flow_id', None)

    def serialize(self) -> ChannelInput:
        return ChannelInput(
            id=self.id,
            task_id=self.task_id,
            flow_id=self.flow_id
        )

    def initialize(self):
        if not self.initialized:
            self.initialized = True
            if self.buffer:
                # always put a END except for constant channel
                # TODO are there any other ways?
                if self.queue_factory is not ConstantQueue:
                    self.buffer.append(END)
                while self.buffer:
                    self.put_nowait(self.buffer.popleft())

    def put_nowait(self, item):
        if self.initialized:
            for q in self.queues:
                q.put_nowait(item)
        else:
            self.buffer.append(item)

    async def put(self, item):
        self.initialize()
        for q in self.queues:
            await q.put(item)

    def create_queue(self) -> LazyAsyncQueue:
        q = LazyAsyncQueue(ch=self, queue_factory=self.queue_factory)
        self.queues.append(q)
        return q


class ConstantChannel(Channel):
    """A channel use ConstantQueue as it's fetcher queue factory.
    """

    def __init__(self, **kwargs):
        super().__init__(queue_factory=ConstantQueue, **kwargs)


class Consumer(Fetcher):
    """Consumer is an object used for simultaneously fetching data emitted by multiple channels and output tuples.
    Empty consumer will emit only once. The end of __next__/__anext__ will be triggered if any of it's source channel
    emits a END object. For simplicity, if there is only a single source channel, the output of consumer will not be
    a tuple.
    """

    def __init__(self, *queues: LazyAsyncQueue, **kwargs):
        super().__init__(**kwargs)
        self.queues: List[LazyAsyncQueue] = list(queues)
        assert all(isinstance(q, LazyAsyncQueue) for q in self.queues)
        self.num_emitted = 0

    @property
    def empty(self):
        return len(self.queues) == 0

    @property
    def single(self):
        return len(self.queues) == 1

    def __len__(self):
        return len(self.queues)

    async def get(self):
        # make sure empty inputs only emit once
        if self.empty and self.num_emitted >= 1:
            return END
        values = []
        for q in self.queues:
            value = await q.get()
            if isinstance(value, End):
                return END
            values.append(value)
        self.num_emitted += 1
        # emit single _input without tuple
        res = tuple(values) if not self.single else values[0]
        return res

    def get_nowait(self):
        # make sure empty inputs only emit once
        if self.empty and self.num_emitted >= 1:
            return END
        values = []
        for q in self.queues:
            value = q.get_nowait()
            if isinstance(value, End):
                return END
            values.append(value)
        self.num_emitted += 1
        # emit single _input without tuple
        return tuple(values) if not self.single else values[0]

    @classmethod
    def from_channels(cls, *channels: Sequence[Union[Channel, object]], **kwargs) -> 'Consumer':
        channels = list(channels)
        for i, ch in enumerate(channels):
            if not isinstance(ch, Channel):
                if isinstance(ch, (tuple, list)) and any(isinstance(v, Channel) for v in ch):
                    raise ValueError(f"The _input: {ch} is a list/tuple of channels, "
                                     f"please unwrap it before pass into a Task/Flow")
                else:
                    channels[i] = Channel.value(ch)
            else:
                if ch.initialized:
                    raise ValueError("Can not create consumer from activated Channel, try to create"
                                     " the consumer before _running the flow.")
        queues = [ch.create_queue() for ch in channels]
        return cls(*queues, **kwargs)


def _a(*args: Union[object, Channel]):
    pass


ARGS_SIG = list(inspect.signature(_a).parameters.values())[0]

Output = Union[Sequence[Channel], Channel]


def _b() -> Output:
    pass


OUTPUT_ANNOTATION = inspect.signature(_b).return_annotation



from typing import List, Optional
from uuid import uuid4

from pydantic import BaseModel, Field


def current_timestamp():
    # utc ?
    from datetime import datetime
    return datetime.utcnow().timestamp()


def uuid():
    return str(uuid4())


class Model(BaseModel):
    pass


class SuccessPayload(Model):
    success: bool = True
    error: str = None


class IdsPayload(Model):
    id: List[str] = Field(default_factory=list)


class RunLog(Model):
    id: str
    level: str
    time: float
    message: str
    task_id: str = None
    flow_id: str = None
    taskrun_id: str = None
    flowrun_id: str = None
    agent_id: str = None


class RunLogInput(Model):
    id: str = Field(default_factory=uuid)
    level: str
    time: float = Field(default_factory=current_timestamp)
    message: str = ""
    task_id: str = None
    flow_id: str = None
    taskrun_id: str = None
    flowrun_id: str = None
    agent_id: str = None


class RunLogsInput(Model):
    logs: List[RunLogInput] = Field(default_factory=list)


class GetRunLogsInput(Model):
    id: List[str] = Field(default_factory=list)
    taskrun_id: List[str] = Field(default_factory=list)
    flowrun_id: List[str] = Field(default_factory=list)
    agent_id: List[str] = Field(default_factory=list)
    level: List[str] = Field(default_factory=list)
    before: float = None
    after: float = None


class Agent(Model):
    id: str
    name: str
    labels: List[str]
    address: str
    flowruns: List[str] = Field(default_factory=list)


class AgentInput(Model):
    id: str
    name: str
    labels: List[str]


class Channel(Model):
    id: str
    task_id: str = None
    flow_id: str = None


class ChannelInput(Model):
    id: str
    task_id: Optional[str]
    flow_id: Optional[str]


class Edge(Model):
    channel_id: str
    task_id: str


class EdgeInput(Model):
    channel_id: str
    task_id: str


class Task(Model):
    id: str
    name: str
    full_name: str
    labels: List[str]
    flow_id: str
    output: List[str]  # in database, it stores id
    docstring: str
    context: dict


class TaskInput(Model):
    id: str
    name: str
    full_name: str
    labels: List[str]
    flow_id: str
    output: List[ChannelInput]
    docstring: str
    context: dict


class Flow(Model):
    id: str
    name: str
    full_name: str
    labels: List[str]
    tasks: List[str]  # in db, store id
    edges: List[Edge]
    docstring: str
    serialized_flow: str
    context: dict
    flowruns: List[str] = Field(default_factory=list)


class FlowInput(Model):
    id: str
    name: str
    full_name: str
    labels: List[str]
    tasks: List[TaskInput]
    edges: List[EdgeInput]
    docstring: str
    context: dict
    serialized_flow: str


class GetFlowsInput(Model):
    id: List[str] = Field(default_factory=list)
    name: List[str] = Field(default_factory=list)
    labels: List[str] = Field(default_factory=list)


class State(Model):
    state_type: str
    result: str = None
    message: str = None


class StateInput(Model):
    state_type: str
    result: str = None
    message: str = None


class RunInput(Model):
    pass


class TaskRun(Model):
    id: str
    flowrun_id: str
    agent_id: str = None
    task_id: str
    flow_id: str
    context: dict
    state: State
    start_time: float = None
    end_time: float = None
    last_heartbeat: float = Field(default_factory=current_timestamp)


class TaskRunInput(RunInput):
    id: str
    flowrun_id: str = None
    agent_id: str = None
    task_id: str = None
    flow_id: str = None
    context: dict = None
    state: StateInput = None
    last_heartbeat: float = Field(default_factory=current_timestamp)


class GetTaskRunsInput(Model):
    id: List[str] = Field(default_factory=list)
    flowrun_id: List[str] = Field(default_factory=list)
    agent_id: List[str] = Field(default_factory=list)
    task_id: List[str] = Field(default_factory=list)
    flow_id: List[str] = Field(default_factory=list)
    state_type: List[str] = Field(default_factory=list)
    after: float = None
    before: float = None


class FlowRun(Model):
    id: str
    agent_id: str = None
    flow_id: str
    name: str
    labels: List[str]
    context: dict
    state: State
    start_time: float = None
    end_time: float = None
    last_heartbeat: float = Field(default_factory=current_timestamp)
    taskruns: List[str] = Field(default_factory=list)  # db store id


class FlowRunInput(RunInput):
    id: str
    agent_id: str = None
    flow_id: str = None
    name: str = None
    labels: List[str] = None
    context: dict = None
    state: StateInput = None
    last_heartbeat: float = Field(default_factory=current_timestamp)


class GetFlowRunsInput(Model):
    id: List[str] = Field(default_factory=list)
    agent_id: List[str] = Field(default_factory=list)
    flow_id: List[str] = Field(default_factory=list)
    name: List[str] = Field(default_factory=list)
    labels: List[str] = Field(default_factory=list)
    state_type: List[str] = Field(default_factory=list)
    after: float = None
    before: float = None



import re

from setuptools import setup, find_packages

classifiers = [
    "Development Status :: 3 - Alpha",
    "Operating System :: POSIX",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "License :: OSI Approved :: MIT License",
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
    "Topic :: Software Development :: Libraries",
]

keywords = [
    "dataflow", "workflow", "asyncio"
]


def get_version():
    with open("flowsaber/__init__.py") as f:
        for line in f.readlines():
            m = re.match("__version__ = '([^']+)'", line)
            if m:
                return m.group(1)
        raise IOError("Version information can not found.")


def get_long_description():
    return "A dataflow based workflow framework."


def get_install_requires():
    requirements = []
    with open('requirements.txt') as f:
        for line in f:
            requirements.append(line.strip())
    return requirements


setup(
    name='flowsaber',
    author='bakezq',
    author_email='zhongquan789@gmail.com',
    version=get_version(),
    license='MIT',
    description=get_long_description(),
    long_description=get_long_description(),
    keywords=keywords,
    url='https://github.com/zhqu1148980644/flowsaber',
    packages=find_packages(),
    scripts=["scripts/flowsaber"],
    include_package_data=True,
    package_data={'': ['server/app/graphql_schema/*.graphql']},
    zip_safe=False,
    classifiers=classifiers,
    install_requires=[
        'makefun',
        'cloudpickle',
        'dask',
        'distributed',
        'psutil',
        'aiohttp',
        'pydantic',
        'fire',
    ],
    extras_require={
        'test': [
            'ariadne',
            'uvicorn',
            'motor',
            # test
            'pytest',
            'pytest-cov',
            'httpimport',
            'autodocsumm',
            # 'pysam',
            'matplotlib'
        ],
        'server': [
            'ariadne',
            'uvicorn',
            'motor'
        ]
    },
    python_requires='>=3.7, <4',
)



"""
Borrowed from prefect.executors.dask
"""

import asyncio
import inspect
import logging
import uuid
import weakref
from concurrent.futures import Future
from typing import Union, Callable, Optional, Any

import flowsaber
from flowsaber.utility.utils import import_object


class Executor(object):
    """Async Executor is aimed for running submitted jobs in an asynchronous way. Executor need to implement three
    __aenter__/__aexit__/run async functions for initializing/running/cleaning.
    """

    def __init__(self, **kwargs):
        pass

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    async def run(self, fn, *args, **kwargs):
        raise NotImplementedError


class Local(Executor):
    """Executor run jobs in the main loop of the current thread,
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    async def run(self, fn, *args, **kwargs):
        coro = fn(*args, **kwargs)
        if inspect.iscoroutine(coro):
            return await coro
        return coro


class DaskExecutor(Executor):
    """
    An executor that runs all functions using the `dask.distributed` scheduler.

    Check https://docs.dask.org/en/latest/setup.html for all kinds of cluster types.

    By default a temporary `distributed.LocalCluster` is created (and
    subsequently torn down) within the `start_loop()` contextmanager. To use a
    different cluster class (e.g.
    [`dask_kubernetes.KubeCluster`](https://kubernetes.dask.org/)), you can
    specify `cluster_class`/`cluster_kwargs`.

    Alternatively, if you already have a dask cluster _running, you can provide
    the address of the scheduler via the `address` kwarg.

    Note that if you have tasks with tags of the form `"dask-accum_resource:KEY=NUM"`
    they will be parsed and passed as
    [Worker Resources](https://distributed.dask.org/en/latest/resources.html)
    of the form `{"KEY": float(NUM)}` to the Dask TaskScheduler.

    Args:
        - address (string, optional): address of a currently _running dask
            scheduler; if one is not provided, a temporary cluster will be
            created in `executor.start_loop()`.  Defaults to `None`.
        - cluster_class (string or callable, optional): the cluster class to use
            when creating a temporary dask cluster. Can be either the full
            class name (e.g. `"distributed.LocalCluster"`), or the class itself.
        - cluster_kwargs (dict, optional): addtional kwargs to pass to the
           `cluster_class` when creating a temporary dask cluster.
        - adapt_kwargs (dict, optional): additional kwargs to pass to `cluster.adapt`
            when creating a temporary dask cluster. Note that adaptive scaling
            is only enabled if `adapt_kwargs` are provided.
        - client_kwargs (dict, optional): additional kwargs to use when creating a
            [`dask.distributed.Client`](https://distributed.dask.org/en/latest/api.html#client).
        - debug (bool, optional): When _running with a local cluster, setting
            `debug=True` will increase dask's logging level, providing
            potentially useful debug info. Defaults to the `debug` value in
            your Prefect configuration.

    Examples:

    Using a temporary local dask cluster:

    ```python
    executor = DaskExecutor()
    ```

    Using a temporary cluster _running elsewhere. Any Dask cluster class should
    work, here we use [dask-cloudprovider](https://cloudprovider.dask.org):

    ```python
    executor = DaskExecutor(
        cluster_class="dask_cloudprovider.FargateCluster",
        cluster_kwargs={
            "image": "prefecthq/prefect:latest",
            "n_workers": 5,
            ...
        },
    )
    ```

    Connecting to an existing dask cluster

    ```python
    executor = DaskExecutor(address="192.0.2.255:8786")
    ```
    """

    def __init__(
            self,
            address: str = None,
            cluster_class: Union[str, Callable] = None,
            cluster_kwargs: dict = None,
            adapt_kwargs: dict = None,
            client_kwargs: dict = None,
            debug: bool = False,
            **kwargs
    ):
        super().__init__(**kwargs)
        if address is not None:
            if cluster_class is not None or cluster_kwargs is not None:
                raise ValueError(
                    "Cannot specify both `address` and `cluster_class`/`cluster_kwargs`"
                )
        from distributed import Client
        from distributed.deploy.local import LocalCluster
        if isinstance(cluster_class, str):
            cluster_class = import_object(cluster_class)
        elif not cluster_class:
            cluster_class = LocalCluster

        self.cluster_class = cluster_class

        self.cluster_kwargs = {} if not cluster_kwargs else cluster_kwargs.copy()
        if cluster_class == LocalCluster:
            self.cluster_kwargs.setdefault(
                'silence_logs', logging.CRITICAL if not debug else logging.WARNING
            )

        self.adapt_kwargs = {} if not adapt_kwargs else adapt_kwargs.copy()
        self.client_kwargs = {} if not client_kwargs else client_kwargs.copy()
        self.client_kwargs.setdefault('set_as_default', False)
        self.client: Optional[Client] = None
        self.cluster: Optional[str, LocalCluster] = None

        self.address = address
        self._futures = None
        self._should_run_event = None
        self._watch_dask_events_task = None

    async def __aenter__(self):
        from distributed import Client
        if self.address is None:
            self.cluster = self.cluster_class(**self.cluster_kwargs, asynchronous=True)
            await self.cluster.__aenter__()
            if self.adapt_kwargs:
                self.cluster.adapt(**self.adapt_kwargs)
        else:
            self.cluster = self.address
        self.client = Client(self.cluster, **self.client_kwargs, asynchronous=True)
        await self.client.__aenter__()
        self._pre_start_yield()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._post_start_yield()
        if not isinstance(self.cluster, str):
            await self.cluster.__aexit__(exc_type, exc_val, exc_tb)
        await self.client.__aexit__(exc_type, exc_val, exc_tb)

    async def run(
            self, fn: Callable, *args: Any, extra_context: dict = None, **kwargs: Any
    ) -> Future:
        """
        Submit a function to the executor for execution. Returns a Future object.

        Args:
            - fn (Callable): function that is being submitted for execution
            - *args (Any): arguments to be passed to `fn`
            - extra_context (dict, optional): an optional dictionary with extra information
                about the submitted task
            - **kwargs (Any): keyword arguments to be passed to `fn`

        Returns:
            - Future: a Future-like object that represents the computation of `fn(*args, **kwargs)`
        """
        if self.client is None:
            raise ValueError("This executor has not been started.")

        kwargs.update(self._prep_dask_kwargs(extra_context))
        if self._should_run_event is None:
            fut = self.client.submit(fn, *args, **kwargs)
        else:
            fut = self.client.submit(
                self._maybe_run, self._should_run_event.name, fn, *args, **kwargs
            )
        res = await fut
        return res

    async def _watch_dask_events(self) -> None:
        scheduler_comm = None
        comm = None
        from distributed.core import rpc

        try:
            scheduler_comm = rpc(
                self.client.scheduler.address,  # type: ignore
                connection_args=self.client.security.get_connection_args("client"),  # type: ignore
            )
            # due to a bug in distributed's inproc comms, letting cancellation
            # bubble up here will kill the listener. wrap with a shield to
            # prevent that.
            comm = await asyncio.shield(scheduler_comm.live_comm())
            await comm.write({"op": "subscribe_worker_status"})
            _ = await comm.read()
            while True:
                try:
                    msgs = await comm.read()
                except OSError:
                    break
                for op, msg in msgs:
                    if op == "add":
                        for worker in msg.get("workers", ()):
                            flowsaber.context.logger.debug("Worker %s added", worker)
                    elif op == "remove":
                        flowsaber.context.logger.debug("Worker %s removed", msg)
        except asyncio.CancelledError:
            pass
        except Exception:
            flowsaber.context.logger.debug(
                "Failure while watching dask worker events", exc_info=True
            )
        finally:
            if comm is not None:
                try:
                    await comm.close()
                except Exception:
                    pass
            if scheduler_comm is not None:
                scheduler_comm.close_rpc()

    def _pre_start_yield(self) -> None:
        from distributed import Event

        is_inproc = self.client.scheduler.address.startswith("inproc")  # type: ignore
        if self.address is not None or is_inproc:
            self._futures = weakref.WeakSet()
            self._should_run_event = Event(
                f"flowsaber-{uuid.uuid4().hex}", client=self.client
            )
            self._should_run_event.set()

        self._watch_dask_events_task = asyncio.run_coroutine_threadsafe(
            self._watch_dask_events(), self.client.loop.asyncio_loop  # type: ignore
        )

    def _post_start_yield(self) -> None:
        from distributed import wait

        if self._watch_dask_events_task is not None:
            try:
                self._watch_dask_events_task.cancel()
            except Exception:
                pass
            self._watch_dask_events_task = None

        if self._should_run_event is not None:
            # Multipart cleanup, ignoring exceptions in each stage
            # 1.) Stop pending tasks from starting
            try:
                self._should_run_event.clear()
            except Exception:
                pass
            # 2.) Wait for all _running tasks to complete
            try:
                futures = [f for f in list(self._futures) if not f.done()]  # type: ignore
                if futures:
                    flowsaber.context.logger.info(
                        "Stopping executor, waiting for %d active tasks to complete",
                        len(futures),
                    )
                    wait(futures)
            except Exception:
                pass
        self._should_run_event = None
        self._futures = None

    def _prep_dask_kwargs(self, extra_context: dict = None) -> dict:
        if extra_context is None:
            extra_context = {}

        dask_kwargs = {"pure": False}  # type: dict

        # set a key for the dask scheduler UI
        key = self._make_task_key(**extra_context)
        if key is not None:
            dask_kwargs["key"] = key

        # infer from context if dask resources are being utilized
        task_tags = extra_context.get("task_tags", [])
        dask_resource_tags = [
            tag for tag in task_tags if tag.lower().startswith("dask-accum_resource")
        ]
        if dask_resource_tags:
            resources = {}
            for tag in dask_resource_tags:
                prefix, val = tag.split("=")
                resources.update({prefix.split(":")[1]: float(val)})
            dask_kwargs.update(resources=resources)

        return dask_kwargs

    def __getstate__(self) -> dict:
        state = self.__dict__.copy()
        state.update({
            k: None
            for k in ["client", "_futures", "_should_run_event", "_watch_dask_events_task"]
        })
        return state

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)

    @staticmethod
    def _make_task_key(
            task_name: str = "", task_index: int = None, **kwargs: Any
    ) -> Optional[str]:
        """A helper for generating a dask task key from field set in `extra_context`"""
        if task_name:
            suffix = uuid.uuid4().hex
            if task_index is not None:
                return f"{task_name}-{task_index}-{suffix}"
            return f"{task_name}-{suffix}"
        return None

    @staticmethod
    def _maybe_run(event_name: str, fn: Callable, *args: Any, **kwargs: Any) -> Any:
        """Check if the task should run against a `distributed.Event` before
        starting the task. This offers stronger guarantees than distributed's
        current cancellation mechanism, which only cancels pending tasks."""
        import dask
        from distributed import Event, get_client

        try:
            # Explicitly pass in the timeout from dask's config_dict. Some versions of
            # distributed hardcode this rather than using the value from the
            # config_dict.  Can be removed once we bump our min requirements for
            # distributed to >= 2.31.0.
            timeout = dask.config.get("distributed.comm.timeouts.connect")
            event = Event(event_name, client=get_client(timeout=timeout))
            should_run = event.is_set()
        except Exception:
            # Failure to create an event is usually due to connection errors. These
            # are either due to flaky behavior in distributed's comms under high
            # loads, or due to the scheduler shutting down. Either way, the safest
            # course here is to assume we *should* run the task still. If we guess
            # wrong, we're either doing a bit of unnecessary work, or the cluster
            # is shutting down and the task will be cancelled anyway.
            should_run = True

        if should_run:
            return fn(*args, **kwargs)


def get_executor(executor_type: str = 'dask', *args, **kwargs):
    executors = {
        'local': Local,
        'dask': DaskExecutor
    }
    if executor_type not in executors:
        raise ValueError(f"{executor_type} not supported, please choose one of {executors.keys()}")
    return executors[executor_type](*args, **kwargs)



"""
"""
from typing import Any, Optional


class State(object):
    """State represents status of flowrun/taskrun.

    State flows and state hierarchy:

        FlowRun:
            Scheduled Pending Running Done
        TaskRun:
            Pending Running Retrying Running Done

        Done
            Success
                Cached
                Skip
            Failure
                Drop
    """

    def __init__(self,
                 state_type: str = None,
                 result: Any = None,
                 message: str = None):
        self.state_type: str = type(self).__name__
        self.result: Optional[Any] = result or ""
        self.message: str = message or ""

    def to_dict(self) -> dict:
        res = {
            'state_type': self.__dict__['state_type'],
            'result': None,
            'message': self.__dict__['message']
        }
        return res

    @classmethod
    def from_dict(cls, state_dict) -> "State":
        state_cls = globals().get(state_dict['state_type'])
        assert issubclass(state_cls, State)
        return state_cls(**state_dict)

    def __repr__(self):
        message = f":{self.message}" if self.message else ""
        return f"<{self.state_type}>({message})"

    @classmethod
    def copy(cls, state: 'State'):
        # TODO, use copy or construction?
        from copy import copy
        # TODO carefully handle inheritance
        # if not issubclass(cls, type(state)):
        #     raise ValueError(f"The copy source's class: {state} must be a supper class of {cls}")
        new = cls()
        new.__dict__ = copy(state.__dict__)
        new.state_type = cls.__name__
        return new


class Scheduled(State):
    pass


class Pending(State):
    pass


class Retrying(Pending):
    """This state comes from Pending, means the task is waiting for rerun due to retry's waiting time"""
    pass


class Running(State):
    pass


class Done(State):
    """Represent the end state of a task run, should not be directly used. Use Success/Failure instead."""
    pass


class Success(Done):
    pass


class Failure(Done):
    """Means some Exception has been raised."""

    def __init__(self, trace_back=None, **kwargs):
        super().__init__(**kwargs)
        self.trace_back = trace_back


class Cached(Success):
    """The result of the _input is cached."""
    pass


class Skip(Success):
    """This state means this _output should be skipped and directly send to the _output channel"""
    pass


class Drop(Done):
    """This state means the _output should be dropped and will not be passed to the _output channel.
    Usually this is caused by settled skip on error option in task.config_dict"""
    pass


class Cancelling(State):
    pass


class Cancelled(Failure):
    pass
